import {PublisherAPI} from './api/PublisherAPI';
import {AD_TYPE} from './common/AD_TYPE';
import {DEFAULT_MIMETYPES} from './common/DEFAULT_MIMETYPES';
import {EVENT} from './events/EVENT';
import {SCHEMA} from './model/schemes/SCHEMA';
import {FacadeBase} from './facades/FacadeBase';
import {Debug} from './utils/Debug';
import {extend} from './utils/extend';
import {loadHomadUtils} from './utils/loadHomadUtils';
import {HomadXHRURLHandler} from './controller/urlhandlers/HomadXHRURLHandler';
import { AdReinsertionModel } from './model/AdReinsertionModel';
import { JSONPURLHandler } from './controller/urlhandlers/JSONPURLHandler';
import { AD_REQUEST_METHOD } from './common/AD_REQUEST_METHOD';

let _playerProxy = null;

const FACADE_VERSION = '3.0';
const SUPPORTED_FACADE_VERSIONS = ['3.0'];

class SmartClientCore extends Debug {

	connect(playerProxy) {

		if (playerProxy) {

			if (this._hasValidHandshake(playerProxy) && this._hasValidFacadeImplementation(playerProxy)) {
				_playerProxy = playerProxy;
				return this;
			}
		}

		this.debug.critical('ConnectError: Calling connect() without player facade reference.');
		return null;
	}

	_hasValidHandshake(playerProxy) {
		if (playerProxy.handshakeVersion) {

			let playerProxyVersion = playerProxy.handshakeVersion(FACADE_VERSION);

			// good version
			if (SUPPORTED_FACADE_VERSIONS.indexOf(playerProxyVersion) > -1) {
				return true;
			}

			this.debug.critical('ConnectError: Handshake failed - Expected Facade version ' + FACADE_VERSION + ' received ' + playerProxyVersion);
			return false;
		}

		this.debug.critical('ConnectError: Handshake failed - Expected Facade version ' + FACADE_VERSION + '. Could not call handshakeVersion(), not implemented?');
		return false;
	}

	_hasValidFacadeImplementation(playerProxy) {

		const playerProxyAPI = Object.keys(playerProxy).concat(Object.getOwnPropertyNames(Object.getPrototypeOf(playerProxy)));
		const facadeBaseAPI = Object.getOwnPropertyNames(Object.getPrototypeOf(new FacadeBase()));

		// not implemented functions or properties will show up in the array
		const apiDifference = facadeBaseAPI.filter((apiFunction) => {
			return !playerProxyAPI.includes(apiFunction);
		});

		if (apiDifference.length > 0) {
			this.debug.critical('ConnectError: Incomplete Facade interface implementation. Expected implementation of following function(s) or properties:', apiDifference.join(','));
			return false;
		}

		return true;
	}

	setup(config) {
		return new Promise((resolve, reject) => {

			if (!_playerProxy) {
				reject(null);
			}
			else if (SCHEMA.validate(config, SCHEMA.GLOBAL_CONFIG)) {
				
				// Note: The publisher might provide just a subset of the global config.
				// This makes sure we can access all global settings, especially the defaults like timeouts.
				let configWithDefaults = extend(SCHEMA.create(SCHEMA.GLOBAL_CONFIG), config);
				
				// CE-HTML (HBB TV): Publishers should use JSONP prior XHR requests.
				if (configWithDefaults.adRequestMethod === AD_REQUEST_METHOD.JSONP) {
					configWithDefaults.urlhandler = JSONPURLHandler;
					
					// disable HOMAD as it uses XHR
					if (configWithDefaults.adReinsertion.homad.enabled) {
						this.debug.warn('[Core] Ad tags get loaded as <script> tags (JSON-P). This disables AdReinsertion which works with XHR requests only.');
					}
					configWithDefaults.adReinsertion.homad.enabled = false;
				}
				
				// Initialize the core API
				let api = new PublisherAPI(_playerProxy, configWithDefaults);
				
				// Load HOMAD
				if (configWithDefaults.adReinsertion.homad.enabled && configWithDefaults.adReinsertion.homad.setup) {
					const homadSetup = configWithDefaults.adReinsertion.homad.setup;
					loadHomadUtils(
						homadSetup.globalConfig,
						homadSetup.clientConfig,	// TODO: URL vs. JSON object
						{
							player: _playerProxy,
							publisherAPI: api
						},
						(homadUtils) => {
							this.debug.info('HOMAD ready, received HOMAD Utils.', homadUtils);
							configWithDefaults.urlhandler = HomadXHRURLHandler;
							configWithDefaults.adReinsertion.homad.utils = homadUtils;
							
							resolve(api);
						},
						(failure) => {
							
							if (!configWithDefaults.adReinsertion.homad.utils) {
								this.debug.warn('HOMAD Initialization Error.', failure);
								
								reject(null);
							}
							else {
								
								// This branch might be requested in case of
								//  - MediaFile errors
								//  - checks & balances detected any irregularities
								this.debug.warn('HOMAD Penalty during ad playback. Request restart?');
								
								// TODO: Consult SvM and request runtime-penalty scenarios.
								/* istanbul ignore else: Ask SvM about possible use-cases. */
								if (api.slot && api.slot.currentAd) {
									api.slot.currentAd.adReinsertion.enable(AdReinsertionModel.REASON.XHR_UNEXPECTED_STATUS);
								}
							}
						}
					);
				}
				else {
					resolve(api);
				}
				
			}
			else {
				reject(null);
			}
		});
	}

	/**
	 * @returns {Object} List of event types as ENUMs
	 */
	get EVENT() {
		return EVENT;
	}

	/**
	 * @returns {Array} List of event types
	 */
	get EVENTS() {
		return Object.values(EVENT);
	}

	get DEFAULT_MIMETYPES() {
		return DEFAULT_MIMETYPES;
	}

	get SCHEMA() {
		return SCHEMA;
	}
	
	/**
	 * @returns {string} Ad type
	 */
	get AD_TYPE() {
		return AD_TYPE;
	}
	
	/**
	 * @returns {string} Ad request method
	 */
	get AD_REQUEST_METHOD() {
		return AD_REQUEST_METHOD;
	}
	
	/**
	 * @returns {Debug} Debug log object
	 */
	get DEBUG() {
		return this.debug;
	}

	/**
	 * @returns {string} Facade version string
	 */
	get supportedFacadeVersion() {
		return FACADE_VERSION;
	}

	/**
	 * @returns {string} smartclientcore version string
	 */
	get version() {
		return '5.0.2';
	}
}

const smartclientcore = new SmartClientCore();

/* istanbul ignore else: There is no fallback object. */
if (typeof window !== 'undefined') {
	window['smartclientcore'] = smartclientcore;
}

/* istanbul ignore else: We cannot bind the core to any other module. */
if (typeof exports === 'object') {
	module.exports = smartclientcore;
}
